"""
    Created on 2012-07-21
    @author: jldupont
"""

class ExcInfo(Exception): pass

class ExcWarn(Exception): pass

class ExcErr(Exception): pass

class ExcCrit(Exception): pass

class ExcQuit(Exception): pass



"""
    Created on 2012-01-27
    @author: jldupont
"""
import json, sys, re, os
import types
import logging
import importlib

from jlddk.tools_logging import setloglevel, enable_duplicates_filter, setup_syslog

class BrokenPipe(Exception): pass


def process_command_line(parser):
    parser.add_argument('-lc',  dest="logconfig", type=str,  help="Logging configuration file", choices=["debug", "info", "warning", "error"])
    parser.add_argument('-ll',  dest='log_level',     type=str,            help="Log Level", default="info", choices=["debug", "info", "warning", "error"])
    parser.add_argument('-kdf', dest='log_keepdup',   action="store_true", help="Keep duplicate log entries", default=False)
    parser.add_argument('-lsl', dest='log_syslog',    action="store_true", help="Enable syslog", default=False)
    args=dnorm(vars(parser.parse_args()))
    
    if (not args["log_keepdup"]):
        enable_duplicates_filter()
        
    if (args["log_syslog"]):
        setup_syslog()
        
    setloglevel(args["log_level"])
    info_dump(args, 20)
    return args

def raise_if_not_ok(code, msg):
    if not code.startswith("ok"):
        raise Exception(msg)


def _echo(*pargs, **kargs):
    print pargs, kargs

def prepare_callable_from_string(st):
    """
    mod.fnc
    """
    mod, fnc=os.path.splitext(st)
    return prepare_callable(mod, fnc.strip("."))

def prepare_mod(module_name):
    try:
        mod=importlib.import_module(module_name)
    except:
        raise Exception("Can't import module '%s'" % module_name)
    return mod


def prepare_callable(module_name, function_name):
    try:
        mod=importlib.import_module(module_name)
    except:
        raise Exception("Can't import module '%s'" % module_name)
        
    try:
        fnc=getattr(mod, function_name)
    except:
        raise Exception("Module '%s' doesn't have a '%s' function" % (mod, function_name))
    
    if not callable(fnc):
        raise Exception("Can't call 'run' function of callable '%S'" % module_name)
    
    return (mod, fnc)


def call(module_name, function_name, *pargs, **kargs):
    """
    >>> call("tools_sys", "_echo", "hello", key="value")
    ('hello',) {'key': 'value'}
    """
    try:
        mod=importlib.import_module(module_name)
    except:
        raise Exception("Can't import module '%s'" % module_name)
        
    try:
        fnc=getattr(mod, function_name)
    except:
        raise Exception("Module '%s' doesn't have a 'run' function" % mod)
    
    if not callable(fnc):
        raise Exception("Can't call 'run' function of callable '%S'" % module_name)
    
    return fnc(*pargs, **kargs)
    
    
def dnorm(d):
    """
    Normalize dictionary
    
    >>> dnorm({"SoMeKeY":"  spaces  "})
    {'somekey': 'spaces'}
    """
    r={}
    for e in d:
        try:    r[e.lower()]=d[e].strip()
        except: r[e.lower()]=d[e]
    return r


def dstrip(d):
    """
    Strip each element in a dict
    
    >>> dstrip({"e1":" v1", "e2":" v2 "})
    {'e1': 'v1', 'e2': 'v2'}
    """
    for e in d:
        try:    d[e]=d[e].strip()
        except: pass
    return d


def info_dump(d, align):
    fmt="%-"+str(align)+"s : %s"

    if type(d)==types.DictionaryType:        
        for key in d:
            logging.info(fmt % (key, d[key]))
            
    if type(d)==types.ListType:
        for el in d:
            key, value=el
            logging.info(fmt % (key, value))

def stdoutjs(jso):
    try:
        o=json.dumps(jso)
        sys.stdout.write(o+"\n")
    except:
        raise BrokenPipe()

def stdout(s):
    try:
        sys.stdout.write(s+"\n")
    except:
        raise BrokenPipe()

def stdout_flush():
    try:
        sys.stdout.flush()
    except:
        raise BrokenPipe()

def json_string(o):
    try:
        return json.dumps(o)
    except:
        return ""

def versa_filter(regexes, inp, filterin=True):
    """
    >>> liste=["in.zip", "out.txt", "out2.txt", "in2.zip"]
    >>> regexes=[r'^([a-zA-Z0-9\-_]+)\.zip$',]
    >>> print versa_filter(regexes, liste)
    ['in.zip', 'in2.zip']
    >>> print versa_filter(regexes, liste, filterin=False)
    ['out.txt', 'out2.txt']
    >>> print versa_filter(regexes, "input.zip")
    input.zip
    """
    if regexes is None:
        return inp
    
    just_one_element=False
    if type(inp)!=types.ListType:
        inp=[inp]
        just_one_element=True
    
    rgs=[]
    for regex in regexes:
        rgs.append(re.compile(regex))
    
    result=[]
    for regex in rgs:
        r=filter(filterfunc(regex, filterin), inp)
        result.extend(r)
    
    if just_one_element:
        return result[0]
    
    return result

def filterfunc(regex, filterin=True):
    """
    >>> import re
    >>> regex=re.compile(r'^([a-zA-Z0-9\-_]+)\.zip$')
    >>> liste=["in.zip", "out.txt", "in2.zip", "out2.txt"]
    >>> result=filter(filterfunc(regex), liste)
    >>> print result
    ['in.zip', 'in2.zip']
    >>> result2=filter(filterfunc(regex, filterin=False), liste)
    >>> print result2
    ['out.txt', 'out2.txt']
    """
    def fn(el):
        try:
            groups=regex.match(el).groups()
            return len(groups[0]) > 0 and filterin
        except:
            return not filterin
        
    return fn


if __name__=="__main__":
    import doctest
    doctest.testmod()
    


"""
    Created on 2012-01-27
    @author: jldupont
"""
import types
from pyfnc import patterned, pattern, coroutine


@pattern(dict, str, "up", bool, types.FunctionType)
def doOnTransition_up(ctx, param, _, current_state, func):
    
    previous_state=ctx.get(param, None)
    ctx[param]=current_state
    
    if (not previous_state or previous_state is None) and current_state:
        func()
    

@pattern(dict, str, "down", bool, types.FunctionType)
def doOnTransition_down(ctx, param, _, current_state, func):
    
    previous_state=ctx.get(param, None)
    ctx[param]=current_state
    
    if (previous_state or previous_state is None) and not current_state:
        func()

@pattern(dict, any, any, bool, None)
def doOnTransition_none(ctx, param, _, current_state, func):
    ctx[param]=current_state


@patterned
def doOnTransition(ctx, param, tr_type, current_state, func):
    """
    >>> import sys
    >>> fn=lambda: sys.stdout.write("test")
    >>> ctx={}
    >>> doOnTransition(ctx, "param", "up", False, fn)
    >>> doOnTransition(ctx, "param", "up", True, fn)
    test
    >>> doOnTransition(ctx, "param", "up", True, fn)
    >>> doOnTransition(ctx, "param", "up", True, fn)
    >>> doOnTransition(ctx, "param", "up", False, fn)
    >>> doOnTransition(ctx, "param", "up", True, fn)
    test
    """
    
@pattern(None, None)
def is_trans_NN(_previous, _current):
    return ('nop', None)

@pattern(None, bool)
def is_trans_NB(_previous, current):
    return ('tr', "up" if current is True else "down")
    
@pattern(False, False)
def is_trans_FF(_previous, _current):
    return ("nop", None)

@pattern(True, True)
def is_trans_TT(_previous, _current):
    return ("nop", None)
    
@pattern(True, False)
def is_trans_TF(_previous, _current):
    return ("tr", "down")

@pattern(False, True)
def is_trans_FT(_previous, _current):
    return ("tr", "up")
    
@pattern(None, any)
def is_trans_NS(_previous, _current):
    return ("tr", "up")
    
@pattern(any, any)
def is_trans_SS(previous, current):
    if previous!=current:
        return ("tr", "ch")
    return ("nop", None)
    
@patterned
def is_trans(previous, current):
    """
    >>> is_trans(True, True)
    ('nop', None)
    >>> is_trans(False, False)
    ('nop', None)
    >>> is_trans(True, False)
    ('tr', 'down')
    >>> is_trans(False, True)
    ('tr', 'up')
    >>> is_trans(None, None)
    ('nop', None)
    >>> is_trans(None, True)
    ('tr', 'up')
    >>> is_trans(None, False)
    ('tr', 'down')
    >>> is_trans(None, "string1")
    ('tr', 'up')
    >>> is_trans("string1", "string2")
    ('tr', 'ch')
    >>> is_trans("string1", "string1")
    ('nop', None)
    >>> is_trans(None, 666)
    ('tr', 'up')
    >>> is_trans(666, 777)
    ('tr', 'ch')
    >>> is_trans(666, 666)
    ('nop', None)
    """

@coroutine
def transition_manager(ctx):
    """
    ctx[param]={ "state": state, "up": fnc, "down": fnc}
    
    >>> import logging
    >>> fu=lambda:logging.warning("up")
    >>> fd=lambda:logging.warning("down")
    >>> ctx={"test":{"up": fu, "down":fd}}
    >>> tm=transition_manager(ctx)
    >>> tm.send(("test", True))   ### warning output 'up'
    >>> tm.send(("test", True))
    >>> tm.send(("test", False))  ### warning output 'down'
    >>> tm.send(("test", False))
    """    
    while True:
        param, current_state=(yield)
        cparam=ctx.get(param, {})
        
        previous_state=cparam.get("previous", None)
        maybe_tr, maybe_dir=is_trans(previous_state, current_state)
        cparam["previous"]=current_state
        ctx[param]=cparam
        
        if maybe_tr=="tr":
            fnc=cparam.get(maybe_dir, None)
            if fnc is not None:
                try:
                    fnc(current_state)
                except:
                    fnc()
        

def simple_transition_manager(ctx, param, current):
    """
    >>> ctx={}
    >>> simple_transition_manager(ctx, "p1", 666)
    ('tr', 'up')
    >>> simple_transition_manager(ctx, "p1", 666)
    ('nop', None)
    >>> simple_transition_manager(ctx, "p1", 777)
    ('tr', 'ch')
    >>> simple_transition_manager(ctx, "p1", 777)
    ('nop', None)
    """
    previous=ctx.get(param, None)
    result=is_trans(previous, current)
    ctx[param]=current
    return result
    

@coroutine
def check_transition():
    """
    Reports transitions
    
    >>> lt=check_transition()
    >>> print lt.send(False)
    ('tr', 'down')
    >>> print lt.send(False)
    ('nop', None)
    >>> print lt.send(True)
    ('tr', 'up')
    >>> print lt.send(True)
    ('nop', None)
    >>> print lt.send(False)
    ('tr', 'down')
    >>> print lt.send(False)
    ('nop', None)
    """
    last_status=None
    result=("nop", None)
    
    while True:
        status=(yield result)
        
        if status!=last_status:
            direction="up" if last_status==False else "down"
            last_status=status
            result=("tr", direction)
        else:
            result=("nop", None)


if __name__=="__main__":
    import doctest
    doctest.testmod()



"""
    Created on 2012-01-27
    @author: jldupont
"""
import logging, sys, json, os
from time import sleep
from tools_os import mkdir_p, get_root_files, file_contents
from tools_os import rm, can_write, resolve_path, move
from tools_func import check_transition

def stdout(jo):
    sys.stdout.write(json.dumps(jo)+"\n")

class BrokenPipe(Exception): pass


def run(source_path=None, move_path=None, check_path=None, 
        batch_size=5, 
        polling_interval=None, enable_delete=False
        ,**_):

    if check_path is not None:
        ct=check_transition()

    if enable_delete and move_path is not None:
        raise Exception("Options '-mp' and '-d' are mutually exclusive")
        
    code, rp=resolve_path(source_path)
    if not code.startswith("ok"):
        raise Exception("can't resolve source path '%s'" % source_path)
    source_path=rp
    
    if move_path is not None:
        code, rp=resolve_path(move_path)
        if not code.startswith("ok"):
            raise Exception("can't resolve 'move_path' '%s'" % move_path)
        move_path=rp

        logging.info("Creating (if necessary) 'move' path: %s" % move_path)
        code, msg=mkdir_p(move_path)
        if not code.startswith("ok"):
            raise Exception("Can't create move path '%s': %s" % (move_path, str(msg)))
            
        logging.info("Checking if 'move' directory is writable")
        code, msg=can_write(move_path)
        if not code.startswith("ok"):
            raise Exception("Can't write to 'move' directory")
            
    to_skip=[]
    logging.info("Process pid: %s" % os.getpid())
    ppid=os.getppid()
    logging.info("Parent pid : %s" % ppid)
    logging.info("Starting loop...")
    while True:
        if os.getppid()!=ppid:
            logging.warning("Parent terminated... exiting")
            break
        
        if check_path is not None:
            try:    exists=os.path.exists(check_path)
            except: exists=False
            
            maybe_tr, _=ct.send(exists)
            if maybe_tr=="tr" and exists:
                logging.info("Check path: passed")
            if maybe_tr=="tr" and not exists:
                logging.info("Check path: failed - skipping")
        else:
            ## fake 'exists'
            exists=True

        if exists:        
            code, files=get_root_files(source_path)
            if not code.startswith("ok"):
                logging.error("Can't get root files from %s" % source_path)
            else:                
                ###############################################################
                files=files[:batch_size]
                try:
                    for src_file in files:
                        
                        if src_file in to_skip:
                            continue
                        
                        code, _=can_write(src_file)
                        if not code.startswith("ok"):
                            to_skip.append(src_file)
                            logging.error("Would not be able to move/delete source file '%s'... skipping streaming" % src_file)
                            continue
        
                        dst_file=None                
                        if move_path is not None:
                            bn=os.path.basename(src_file)
                            dst_file=os.path.join(move_path, bn)
                        
                        code, maybe_error=process(src_file, dst_file, enable_delete)
                        if not code.startswith("ok"):
                            to_skip.append(src_file)
                            logging.warning("Problem processing file '%s': %s" % (src_file, maybe_error))
                except BrokenPipe:
                    raise
                except KeyboardInterrupt:
                    raise
                except Exception, e:
                    logging.error("processing file '%s': %s" % (src_file, str(e)))
                ###############################################################            
        
        
        logging.debug("...sleeping for %s seconds" % polling_interval)
        sleep(polling_interval)


def process(src_file, dst_file, enable_delete):
    """
    1. read file, extract URL
    2. send "start"
    3. send each line
    4. send "end"
    5. move/delete source file
    """
    code, contents=file_contents(src_file)
    if not code.startswith("ok"):
        return ("error", "file/invalid")
    
    try:    
        contents=contents.strip()
        lines=contents.split("\n")
    except:
        return ("error", "data/invalid")
    
    ###############################################
    try:
        stdout({"sp": src_file, "code":"begin"})    
        for line in lines:
            stdout({"code": "line", "line": line})
        stdout({"sp": src_file, "code":"end"})
    except:
        raise BrokenPipe("Broken Pipe")
    ###############################################
    
        
    if enable_delete:
        code, _msg=rm(src_file)
        if not code.startswith("ok"):
            logging.error("Can't delete '%s'" % src_file)
            return ("error", "file/delete")
        return ("ok", None)
    
    ### well then, we need to move the source_file
    code, _=move(src_file, dst_file)
    return (code, "file/move")




"""
    Created on 2012-01-27
    @author: jldupont
"""
import os, logging, glob
from jlddk.tools_os import atomic_write, mkdir_p, file_contents, rmdir
from time import sleep

class ExpOK(Exception): pass
class ExpWarning(Exception): pass

def run(
        path_dest=None
        ,path_src=None
        ,file_output_ext=None
        ,file_input_pattern=None
        ,poll_interval=None
        ,start_of_file=None
        ,delete_source_dir=False
        ,**_
        ):
    
    start_of_file=start_of_file.strip("'\"")
    
    file_output_ext=file_output_ext.strip(".'\"")
    logging.info("Using '%s' as output file extension" % file_output_ext)

    ## SOURCE
    logging.info("Resolving source path: %s" % path_src)
    try:
        apath=os.path.abspath(path_src)
        spath=apath.strip("\"'")
        spath=os.path.expanduser(os.path.expandvars(spath))
    except:
        raise Exception("Can't source resolve path")
    
    logging.info("Resolve source path: %s" % spath)
    logging.info("Creating source, if required...")
    mkdir_p(spath)

    ## DESTINATION
    logging.info("Resolving destination path: %s" % path_dest)
    try:
        apath=os.path.abspath(path_dest)
        dpath=apath.strip("\"'")
        dpath=os.path.expanduser(os.path.expandvars(dpath))
    except:
        raise Exception("Can't resolve destination path")
        
    logging.info("Resolve destination path: %s" % dpath)
    logging.info("Creating destination, if required...")
    mkdir_p(dpath)

    def filter_tilde(path):
        return not path.startswith("~")

    ppid=os.getppid()        
    logging.info("Process pid: %s" % os.getpid())
    logging.info("Parent pid:  %s" % ppid)
    logging.info("Starting loop...")
    while True:
        if os.getppid()!=ppid:
            logging.warning("Parent terminated... exiting")
            break

        try:            
            try:
                dirs=os.listdir(spath)
                logging.debug("Dirs: %s" % len(dirs))
            except:
                raise ExpWarning("Can't get subdirectories of source path...")
            
            try:
                fdirs=filter(filter_tilde, dirs)
                logging.debug("Filtered dirs: %s" % len(fdirs))
            except:
                raise ExpWarning("Can't apply filter for ~dir...")
            
            if len(fdirs)==0:
                logging.debug("No directories to work on")
                raise ExpOK()
            
            wdir=os.path.join(spath, fdirs[0])
            logging.debug("Working on path: %s" % wdir)
            
            try:
                files=glob.glob(wdir+os.path.sep+file_input_pattern)
                files=filter(os.path.isfile, files)
                logging.debug("Got: %s file(s) in selected dir" % len(files))
            except:
                raise ExpWarning("Can't glob files...")
            
            for _file in files:
                logging.info("Processing: '%s'" % _file)
                num_files=process(dpath, _file, start_of_file, file_output_ext)
                logging.info("Progress> processed 1 concatenated file with '%s' files" % num_files)
                
            if len(files)>0:
                logging.info("Progress> processed %s files" % len(files))
                
                if delete_source_dir:
                    logging.info("Deleting source directory: %s" % wdir)
                    code, msg=rmdir(wdir)
                    if not code.startswith("ok"):
                        raise ExpWarning("Can't delete source directory: %s" % str(msg))
                                
        except ExpWarning,e:
            logging.warning(e)
            
        except ExpOK,e:
            pass
        
        logging.debug("Sleeping for %s seconds..." % poll_interval)
        sleep(poll_interval)    
        

def process(dpath, _file, start_of_file, file_output_ext):
    
    fn=os.path.splitext(_file)[0]
    bn=os.path.basename(fn)
    odir=os.path.join(dpath, bn)
    odirtemp=os.path.join(dpath, "~"+bn)
    
    if os.path.exists(odirtemp):
        logging.info("Deleting output temporary directory: %s" % odirtemp)
        code, _msg=rmdir(odirtemp)
        if not code.startswith("ok"):
            logging.warning("Can't remove temporary directory... trying to proceed anyways")   

    if os.path.exists(odir):
        logging.info("Deleting output directory: %s" % odir)
        code, _msg=rmdir(odir)
        if not code.startswith("ok"):
            logging.warning("Can't remove output directory... trying to proceed anyways")   
    
    logging.info("Creating output temporary directory name: %s" % odirtemp)
    code, _=mkdir_p(odirtemp)
    
    if not code.startswith("ok"):
        raise ExpWarning("Can't create temporary output directory: %s" % odirtemp)
    
    #########################
    
    
    try:
        logging.debug("Opening file: %s" % _file)
        fh=open(_file)
    except:
        raise ExpWarning("Can't open file: %s" % _file)

    buf=[]
    index=0
    for line in fh:
        if line.startswith(start_of_file):
            write_file(index, buf, bn, odirtemp, file_output_ext)
            buf=[line,]           
            index=index+1
            continue
        else:
            buf.append(line.strip())
            
    write_file(index, buf, bn, odirtemp, file_output_ext)
    try:
        fh.close()
    except:
        pass
    
    logging.info("Renaming directory: %s => %s" % (odirtemp, odir))
    try:
        os.rename(odirtemp, odir)
    except:
        raise ExpWarning("Can't rename %s => %s" % (odirtemp, odir))
    
    return index
        
        
def write_file(index, buf, bn, dpath, file_output_ext):
    if len(buf)==0:
        return
    
    fn=bn+"_"+str(index)+"."+file_output_ext
    logging.debug("Writing file: %s => %s" % (index, fn))
    
    contents="\n".join(buf)
    
    fp=os.path.join(dpath, fn)
    code, _=atomic_write(fp, contents)
    if not code.startswith("ok"):
        raise ExpWarning("Can't write to: %s" % fp)

    logging.debug("Wrote file: %s" % fp)
    




"""
    Created on 2012-01-27
    @author: jldupont
"""
import logging, sys, os
from tools_sys import prepare_callable


def run(module=None, function=None, function_args=None, 
        no_filter_empty=None
        ,**_):
    
    try:                
        logging.info("Preparing callable '%s' from module '%s'" % (function, module))
        _mod, fnc=prepare_callable(module, function)
    except:
        raise Exception("%s.%s isn't callable..." %(module, function))

    try:                
        init_function_name="%s_init" % function
        logging.info("Preparing init '%s' from module '%s'" % (init_function_name, module))
        _mod, fnc_init=prepare_callable(module, init_function_name)
    except Exception,e:
        logging.info("No or errored init function...: %s" % e)
        fnc_init=None

    if fnc_init is not None:
        try:
            logging.debug("Calling init function...")
            fnc_init(*function_args)
        except:
            raise Exception("Error calling init function")

    ppid=os.getppid()            
    logging.info("Process pid: %s" % os.getpid())
    logging.info("Parent pid : %s" % ppid)
    logging.info("Starting loop...")
    while True:
        if os.getppid()!=ppid:
            logging.warning("Parent terminated... exiting")
            break
        
        try:
            iline=sys.stdin.readline().strip(" \n")
        except:
            raise Exception("Exiting... probably broken pipe")

        if not no_filter_empty:
            if iline=="":
                logging.debug("Skipping empty line...")
                continue

        try:
            oline=fnc(iline, *function_args)
        except KeyboardInterrupt:
            raise
        except Exception, e:
            try:    logging.error("Error processing '%s' : %s" % (iline[:20], str(e)))
            except: pass
            continue

        if oline is None:
            logging.debug("No return value from called function")
            continue
        
        try:
            ostr=str(oline).strip("\n")
        except:
            logging.error("Error trying to stringify return string... continuing")
            continue
            
        try:
            sys.stdout.write(ostr+"\n")
        except:
            raise Exception("Exiting... probably broken pipe")




"""
    Created on 2012-01-27
    @author: jldupont
"""
import logging, sys, json, os
from time import sleep
from tools_os import get_root_files
from tools_os import resolve_path
from tools_func import doOnTransition, transition_manager
from tools_misc import check_if_ok
from tools_func import check_transition
from tools_sys import BrokenPipe

from pyfnc import patterned, pattern, partial

def ilog(path):
    logging.info("Files accessible on path: %s" % path)
    
def wlog(path):
    logging.warning("Can't retrieve files from path: %s" % path)

def stdoutf():
    sys.stdout.flush()
    
def stdoutj(jo):
    try:    sys.stdout.write(json.dumps(jo)+"\n")
    except: 
        raise BrokenPipe("...broken pipe")
    
def stdout(m):
    try:    sys.stdout.write(m+"\n")
    except: 
        raise BrokenPipe("...broken pipe")

def run(primary_path=None, compare_path=None, 
        status_filename=None, check_path=None
        ,just_basename=None
        ,topic_name=None
        ,exts=None
        ,wait_status=None, polling_interval=None
        ,just_zppp=None, just_ppzp=None, just_com=None
        ,**_):

    if check_path is not None:
        ct=check_transition()

    code, primary_path=resolve_path(primary_path)
    if not code.startswith("ok"):
        raise Exception("can't resolve primary path '%s'" % primary_path)
    
    code, compare_path=resolve_path(compare_path)
    if not code.startswith("ok"):
        raise Exception("can't resolve compare path '%s'" % compare_path)
            
    if wait_status:
        status_path=os.path.join(primary_path, status_filename)
        logging.info("Using status file path: %s" % status_path)
    else: 
        status_path=None

    ### context for logging etc.
    ctx={
          "just_zppp": just_zppp
         ,"just_ppzp": just_ppzp
         ,"just_com": just_com
         ,"just_list": just_zppp or just_ppzp or just_com
         ,"pp": primary_path
         ,"zp": compare_path
         ,"sp": status_path
         
         ,"pp_log" :{"up":    partial(ilog, primary_path)
                     ,"down":  partial(wlog, primary_path)
                     }
         ,"zp_log" :{"up":    partial(ilog, compare_path)
                     ,"down":  partial(wlog, compare_path)
                     }
         ,"topic_name": topic_name
         ,"exts": exts
         }

    ctx["tm"]=transition_manager(ctx)
    
    ppid=os.getppid()        
    logging.info("Process pid: %s" % os.getpid())
    logging.info("Parent pid: %s" % ppid)
    logging.info("Starting loop...")
    while True:
        if os.getppid()!=ppid:
            logging.warning("Parent terminated... exiting")
            break
            
        if check_path is not None:
            try:    exists=os.path.exists(check_path)
            except: exists=False
            
            maybe_tr, _=ct.send(exists)
            if maybe_tr=="tr" and exists:
                logging.info("Check path: passed")
            if maybe_tr=="tr" and not exists:
                logging.info("Check path: failed - skipping")
        else:
            ## fake 'exists'
            exists=True

        if exists:            
            code, msg=check_if_ok(status_path, default="ok")
            maybe_process(ctx, code, msg, primary_path, compare_path, just_basename)
        
        logging.debug("...sleeping for %s seconds" % polling_interval)
        sleep(polling_interval)

def filtre(exts):
    def _(path):
        _, ext=os.path.splitext(path)
        ext=ext.lstrip(".")
        return ext in exts
    return _
        


@pattern(dict, "ok", any, str, str, any)
def maybe_process_ok(ctx, _ok, _, primary_path, compare_path, just_basename):
    
    ### log rate limiter helper -- need to update status in context 'ctx'
    doOnTransition(ctx, "status_file.contents", "down", True, None)
    
    codep, primary_files=get_root_files(primary_path, strip_dirname=True)
    codec, compare_files=get_root_files(compare_path, strip_dirname=True)
    
    ### output some log info on transitions
    tm=ctx["tm"]
    tm.send(("pp_log", codep=="ok"))
    tm.send(("zp_log", codec=="ok"))
    
    ### not much to do if either path isn't accessible...
    if not codep.startswith("ok") or not codec.startswith("ok"):
        return
    
    exts=ctx["exts"]
    if exts is not None:
        primary_files=filter(filtre(exts), primary_files)
        compare_files=filter(filtre(exts), compare_files)
    
    if just_basename:
        pfiles=map(os.path.basename, primary_files)
        cfiles=map(os.path.basename, compare_files)
    else:
        pfiles=primary_files
        cfiles=compare_files
    
    try:
        setpf=set(pfiles)
        setcf=set(cfiles)
        common=setpf.intersection(setcf)
        
        
        
        diff={
               "pp":     primary_path
              ,"zp":     compare_path
              ,"pp-zp":  list(setpf-setcf)
              ,"zp-pp":  list(setcf-setpf)
              ,"common": list(common)
              }
        
        topic_name=ctx["topic_name"]
        if topic_name is not None:
            diff["topic"]=topic_name
        
        if ctx["just_list"]:
            doout(ctx, diff)
        else:
            stdoutj(diff)
            stdoutf()
    except Exception, e:
        logging.error("Can't compute diff between paths: %s" % str(e))

@pattern(dict, any, str, any, any, any)
def maybe_process_nok(ctx, _nok, _msg, _x, _y, _bn):
    """
    Rate limit logs
    """
    def wlog():
        logging.warning("Can't retrieve file contents from: %s" % ctx["sp"])
         
    doOnTransition(ctx, "status_file.contents", "down", False, wlog)


@patterned
def maybe_process(ctx, code, msg, primary_path, compare_path, just_basename): pass


def doout(ctx, listes):

    if ctx["just_zppp"]:
        for e in listes["zp-pp"]:
            stdout(e)
            stdoutf()

    if ctx["just_ppzp"]:
        for e in listes["pp-zp"]:
            stdout(e)
            stdoutf()
    
    if ctx["just_com"]:
        for e in listes["common"]:
            stdout(e)
            stdoutf()

    




"""
    Created on 2012-02-08
    @author: jldupont
"""
from time import sleep
import json, logging
from tools_os import file_contents

from pyfnc import patterned, pattern

def debug(*p):
    print p
    return p[0]

def batch(l, size):
    """
    >>> for b in batch([1,2,3,4,5,6], 2): print b
    [1, 2]
    [3, 4]
    [5, 6]
    >>> for b in batch([1], 2): print b
    [1]
    >>> for b in batch([], 2): print b
    """
    index=0
    while True:
        sub=l[index:index+size]
        index=index+size
        if len(sub)==0:
            break
        
        yield sub
        


def check_if_ok(filepath, default="ok"):
    """
    Check for 'ok' string or JSON object with 'ok' code
    
    >>> check_if_ok(None)
    ('ok', None)
    >>> check_if_ok("/tmp/non_existing_path")
    ('error', 'empty file')
    >>> from tools_os import quick_write, rm
    >>> p="/tmp/_jlddk_tools_misc_test"
    >>> quick_write(p, "ok")
    ('ok', '/tmp/_jlddk_tools_misc_test')
    >>> check_if_ok(p)
    ('ok', None)
    >>> quick_write(p, '''{"code": "ok"}''')
    ('ok', '/tmp/_jlddk_tools_misc_test')
    >>> check_if_ok(p)
    ('ok', None)
    >>> quick_write(p, '''{"code": "error"}''')
    ('ok', '/tmp/_jlddk_tools_misc_test')
    >>> check_if_ok(p) # doctest: +ELLIPSIS
    ('error', ...
    >>> rm(p)
    ('ok', '/tmp/_jlddk_tools_misc_test')
    """
    
    if filepath is None:
        return (default, None)
    
    code, maybe_contents=file_contents(filepath)
    return checkok(code, maybe_contents)


@pattern(str, None)
def checkok_empty(_, contents):
    return ("error", "empty file")

@pattern("ok", str)
def checkok_ok(_, contents):
    contents=contents.strip()
    
    if contents.startswith("ok"):
        return ("ok", None)
    
    try:
        jo=json.loads(contents)
        if jo["code"]=="ok":
            return ("ok", None)
    except:
        return ("error", "invalid format")
    
    return ("error", "'ok' code not found")

@pattern(any, any)
def check_nok(code, msg):
    return (code, msg)

@patterned
def checkok(code, maybe_contents): pass


def retry(f, always=True, min_wait=1, max_wait=30, max_retries=10, logmsg=None):
    """
    Retries function 'f' : the function should throw an exception to indicate failure 
    
    :param always: boolean, should always retry true/false
    :param min_wait: int, minimum seconds between tries
    :param max_wait: int, maximum seconds before restarting cycle
    :param max_retries: int, maximum number of retries when 'always==False'
    
    @raise KeyboardInterrupt
    @return f()
    """
    showed_msg=False
    wait=min_wait
    retry_count=max_retries
    while True:
        try:
            return f()
        except KeyboardInterrupt:
            raise
        except Exception, exp:   
            if not showed_msg:
                showed_msg=True
                if logmsg is not None:
                    logging.warning(logmsg)     
            retry_count=max(retry_count-1, 0)
            if not always and retry_count==0:
                raise exp
            
        sleep(wait)
        
        wait=wait*2
        wait=wait if wait<max_wait else min_wait


if __name__=="__main__":
    import doctest
    doctest.testmod()



"""
    Created on 2012-01-19
    @author: jldupont
"""
import os, errno, tempfile, types, shutil
import subprocess

from pyfnc import patterned, pattern, partial


def getsubdirs(path):
    """
    >>> getsubdirs("/tmp")
    """
    try:
        paths=os.listdir(path)
        paths=map(partial(os.path.join, path), paths)
        dirs=filter(os.path.isdir, paths)
        return ("ok", dirs)
    except Exception,e:
        return ("error", e)


def move(src_path, dst_path):
    try:
        os.rename(src_path, dst_path)
        return ("ok", None)
    except:
        return ('error', None)


def file_contents(path):
    """
    Simple "get file contents"
    """
    try:
        fh=open(path, "r")
        return ('ok', fh.read())
    except:
        return ("error", None)
    finally:
        try:
            fh.close()
        except:
            pass

def atomic_write(path, contents, tmppath=None):
    """
    Atomic write to file
    
    Create temporary file and then move/rename to specified path.
    Rename operation in the same filesystem are atomic (at least in Linux).
    
    >>> atomic_write("/tmp/_jlddk_atomic_write", "test!") ## doctest: +ELLIPSIS
    ('ok', ...
    """
    
    dn=tmppath or os.path.dirname(path)
    fd, tfn=tempfile.mkstemp(dir=dn)
    
    try:
        ### part 1: write to temp file
        f=os.fdopen(fd, "w")
        f.write(contents)
        f.close()
    except Exception, e:
        try:    os.close(fd)
        except: pass
        return ("error", "write to temp file: %s" % str(e))        
        
    try:
        ### part 2: rename to specified path
        os.rename(tfn, path)
    except Exception,e:
        return ("error", "rename to path '%s': %s" % (path, e))
        
    return ("ok", tfn)
        
    

def quick_write(path, contents):
    """
    Best effort "write contents to file" 
    
    >>> quick_write("/tmp/QuickWriteTest", "{'param': 'value'}") # doctest:+ELLIPSIS
    ('ok', ...)
    >>> file_contents("/tmp/QuickWriteTest")
    ('ok', "{'param': 'value'}")
    >>> rm('/tmp/QuickWriteTest') # doctest:+ELLIPSIS
    ('ok', ...)
    """
    try:
        code, maybe_rpath=resolve_path(path)
        if not code.startswith("ok"):
            return (code, maybe_rpath) 
        fh=open(maybe_rpath, "w")
        fh.write(contents)
        fh.close()
        return ('ok', path)
    except Exception, e:
        return ("error", str(e))
    finally:
        try:
            fh.close()
        except:
            pass


@pattern(str, None, list)
def filter_files_by_ext_NL(criteria, extlist, files):
    return files

@pattern(str, None, types.TupleType)
def filter_files_by_ext_NA(criteria, extlist, (code, maybe_list)):
    if not code.startswith("ok"):
        return []
    return maybe_list


@pattern(str, list, types.TupleType)
def filter_files_by_ext_tup(criteria, extlist, (code, maybe_list)):
    if not code.startswith("ok"):
        return []
    
    return filter_files_by_ext(criteria, extlist, maybe_list)


@pattern("include", list, list)
def filter_files_by_ext_inc(criteria, extlist, files):
    
    def inf(path):
        _root, ext=os.path.splitext(path)
        return ext in extlist
    
    return filter(inf, files)

@pattern("exclude", list, list)
def filter_files_by_ext_exc(criteria, extlist, files):

    def exf(path):
        _root, ext=os.path.splitext(path)
        return not ext in extlist
    
    return filter(exf, files)

@patterned
def filter_files_by_ext(criteria, extlist, files):
    """
    >>> filter_files_by_ext("include", [".json"], ["a.json", "b.json", "c.py"])
    ['a.json', 'b.json']
    >>> filter_files_by_ext("exclude", [".json"], ["a.json", "b.json", "c.py"])
    ['c.py']
    >>> filter_files_by_ext("include", [".json"], ("ok", ["a.json", "b.json", "c.py"]))
    ['a.json', 'b.json']
    >>> filter_files_by_ext("exclude", [".json"], ("error", None))
    []
    """

    
def get_root_files(src_path, strip_dirname=False):
    """
    Retrieve files from the root of src_path
    >> get_root_files("/tmp")
    
    >> get_root_files("/tmp", strip_dirname=True)
    
    """
    def add_src_path(p):
        return os.path.join(src_path, p)
    
    try:
        liste=os.listdir(src_path)
        liste=map(add_src_path, liste)        
        liste=filter(os.path.isfile, liste)
        
        if strip_dirname:
            liste=map(os.path.basename, liste)
     
        return ("ok", liste)
    except:
        return ("error", None)

def touch(path):
    """
    >>> touch("/tmp/JustTouching")
    ('ok', '/tmp/JustTouching')
    >>> rm('/tmp/JustTouching')
    ('ok', '/tmp/JustTouching')
    """
    fhandle = file(path, 'a')
    try:
        os.utime(path, None)
        return ('ok', path)
    except OSError, exc:
        return ("error", (exc.errno, errno.errorcode[exc.errno]))
    finally:
        fhandle.close()        

def rm(path):
    """
    Silently (i.e. no exception thrown) removes a path if possible
    
    >>> rm("/tmp/NotAFile") ## no need to complain if there is no file
    ('ok', '/tmp/NotAFile')
    """
    try:
        os.remove(path)
        return ('ok', path)
    except OSError, exc:
        if exc.errno==errno.ENOENT:
            return ('ok', path)
        return ("error", (exc.errno, errno.errorcode[exc.errno]))


def rmdir(path):
    try:
        shutil.rmtree(path)
        return ("ok", None)
    except Exception,e:
        return ("error", e)


def can_write(path):
    """
    Checks if the current user (i.e. the script) can delete the given path
    Must check both user & group level permissions
    
    FOR THIS TEST, NEED TO CREATE A DIRECTORY /tmp/_test_root_sipi THROUGH ROOT
    
    >>> p="/tmp/_test_sipi"
    >>> rm(p)
    ('ok', '/tmp/_test_sipi')
    >>> touch(p)
    ('ok', '/tmp/_test_sipi')
    >>> can_write(p)
    ('ok', True)
    >>> rm(p)
    ('ok', '/tmp/_test_sipi')
    >>> can_write("/tmp/_test_root_sipi/")
    ('ok', False)
    """
    try:
        return ("ok", os.access(path, os.W_OK))
    except:
        return ("error", None)



def remove_common_prefix(common_prefix, path):
    """
    >>> remove_common_prefix("/tmp", "/tmp/some_dir/some_file.ext")
    ('ok', '/some_dir/some_file.ext')
    """
    try:
        _head, _sep, tail=path.partition(common_prefix)
        return ("ok", tail)
    except:
        return ("error", path)


def gen_walk(path, max_files=None):
    count=0
    done=False
    for root, _dirs, files in os.walk(path):
        
        for f in files:
            yield os.path.join(root, f)
        
            count=count+1
            if max_files is not None:
                if count==max_files:
                    done=True
                    break
            
        if done: break
        



def safe_listdir(path):
    try:
        return ("ok", os.listdir(path))
    except Exception, e:
        return ("error", e)

def get_path_extension(path):
    """
    >>> get_path_extension("/some/path/config.yaml")
    ('/some/path/config', '.yaml')
    >>> get_path_extension(".bashrc")
    ('.bashrc', '')
    >>> get_path_extension("/some/path/file")
    ('/some/path/file', '')
    """
    return os.path.splitext(path)

def mkdir_p(path):
    """
    Silently (i.e. no exception thrown) makes a directory structure if possible
    
    This function can be found in the 'sipi' package.
    """
    try:
        os.makedirs(path)
        return ('ok', path)
    except OSError, exc:
        if exc.errno == errno.EEXIST:
            return ('ok', path)

        return ('error', (exc.errno, errno.errorcode[exc.errno]))


def resolve_path(path):
    try:
        path=os.path.expandvars(path)
        path=os.path.expanduser(path)
        return ("ok", path)
    except:
        return ("error", None)


def safe_isfile(path):
    try:
        return os.path.isfile(path)
    except:
        return False
    
    
def simple_popen(path_script, env={}):
    new_env=os.environ.copy()
    new_env.update(env)
    return subprocess.Popen(path_script, env=new_env)

    
    

if __name__=="__main__":
    import doctest
    doctest.testmod()
    


"""
    Created on 2012-07-21
    @author: jldupont
"""
import logging
from time import sleep



def init(mqueue, _msg):
    mqueue.append({"topic": "tick"})

def tick(mqueue, _):
    logging.info("test: tick!")
    sleep(1)
    mqueue.append({"topic": "tick"})



__vtable__={"init": init, "tick": tick}    


"""
    Created on 2012-02-17
    @author: jldupont
    
"""
import logging, sys, os 

from tools_sys import BrokenPipe

def run(iformat=None
        ,separator=None
        ,**_
        ):
        
    if separator is None or separator=="":
        separator=" "
        
    num_params=iformat.count("%s")
    logging.info("Expected number of input params: %s" % num_params)
        
    ppid=os.getppid()        
    logging.info("Process pid: %s" % os.getpid())
    logging.info("Parent pid:  %s" % ppid)
    logging.info("Starting loop...")
    while True:
        if os.getppid()!=ppid:
            logging.warning("Parent terminated... exiting")
            break
            
        try:
            iline=sys.stdin.readline().strip()
            if len(iline)==0:
                continue
        except:
            raise Exception("Exiting... probably broken pipe")
        
        params=iline.split(separator)
        
        logging.debug("Params: %s" % params)
        try:
            result=iformat % tuple(params)
        except Exception,e:
            logging.error(e)
            continue
        
        try:
            sys.stdout.write(result+"\n")
        except:
            raise BrokenPipe("broken pipe...")




"""
    Created on 2012-01-27
    @author: jldupont
"""
import logging, sys, json, os, select, time, types

from pyfnc import patterned, pattern

def stdout(jo):
    try:    
        sys.stdout.write(json.dumps(jo)+"\n")
        sys.stdout.flush()
    except:
        raise Exception("Exiting... probably broken pipe")


def run( ctx 
        ,polling_interval=None
        ,**_
        ):

    ctx["pairs"]={}
    ctx["count"]=0

    ppid=os.getppid()
    logging.info("Process pid: %s" % os.getpid())
    logging.info("Parent pid : %s" % ppid)
    logging.info("Starting loop...")    
    while True:
        if os.getppid()!=ppid:
            logging.warning("Parent terminated... exiting")
            break

        ### PASS-THROUGH LOGIC
        ######################
        start_time=time.time()
        while True:
            ir, _w, _e=select.select([sys.stdin], [], [], polling_interval)
            if len(ir):
                iline=sys.stdin.readline()
                oline=process_input(ctx, iline)
                if oline is not None:
                    sys.stdout.write(oline)
                    sys.stdout.flush()
                
            elapsed_time = time.time() - start_time
            if elapsed_time > polling_interval:
                break

        perform_update(ctx)
        ctx["count"]=ctx["count"]+1
        
        logging.debug("...end of cycle #%s" % ctx["count"])
        #### /while
        

def process_input(ctx, iline):
    """
    1- first check if we are dealing with a JSON object
    2- next, check if topic match
    3- and make sure the required fields are present
    --
    4- 
    
    """
    try:
        jso=json.loads(iline)
        topic=jso["topic"]
        assert(topic==ctx["itopic"])
        keyfieldname=ctx["key"]
        valuefieldname=ctx["value"]
        
        ### if we can't extract those, get out!
        key=jso[keyfieldname]
        value=jso[valuefieldname]
    except:
        ### pass-through
        logging.debug("pass-through: %s" % iline)
        return iline
    
    logging.debug("Got: {%s:%s}" % (key, value))
    
    ### at this point, we have everything we need
    ###  to perform our work
    pairs=ctx["pairs"]
    
    entry=pairs.get(key, (0, 0, 0, False))
    _last_update_count, _last_change_count, last_value, _stability=entry
    
    result=handle(ctx, entry, key, value, value==last_value)
    if result is None:
        result=""
        
    return iline+result

#######################################################################################    

@pattern(dict, types.TupleType, types.UnicodeType, any, True)
def handle_nochange(ctx, entry, key, _value, _):
    """
    No change --> just update "last_update_count"
    """
    _last_update_count, last_change_count, last_value, stability=entry
    current_count=ctx["count"]
    entry=(current_count, last_change_count, last_value, stability)
    pairs=ctx["pairs"]
    pairs[key]=entry
    return None
    

@pattern(dict, types.TupleType, types.UnicodeType, any, False)
def handle_change(ctx, entry, key, value, _):
    """
    There is a change...
    """
    _last_update_count, _last_change_count, _last_value, _stability=entry
    current_count=ctx["count"]
    entry=(current_count, current_count, value, False)
    pairs=ctx["pairs"]
    pairs[key]=entry
    return None
    
@patterned
def handle(ctx, entry, key, value, changed): pass


#######################################################################################

def perform_update(ctx):
    """
    check for 'stale' k:v pairs
    
    Use "modular arithmetic" will a sufficiently high circumference
     ensuring that runovers don't happen
    """
    hin=ctx["hin"]
    timeout=ctx["timeout"]
    current_count=ctx["count"]
    
    states=[]
    new={}
    pairs=ctx["pairs"]
    for key in pairs:
        entry=pairs[key]
        last_update_count, last_change_count, last_value, stability=entry

        ### check if stale...
        if (current_count - last_update_count) % 8192 > timeout:
            logging.info("Deleting stale key: %s" % key)
            continue
        
        ### check if stable...
        if (current_count - last_change_count) % 8192 >= hin:
            if not stability:
                logging.info("Stable {%s:%s}" % (key, last_value))
            
            ### take current count as "last_change_count"
            ###  trickery to keep complexity low...
            cc=(current_count - hin) % 8192
            new[key]=(last_update_count, cc, last_value, True)
            states.append((key, last_value, True))  ## stable
            continue
        else:
            states.append((key, last_value, False)) ## !stable
            new[key]=(last_update_count, last_change_count, last_value, False)
            
    ### clean dict without stale entries
    ctx["pairs"]=new
    
    use_kv=ctx["use_kv"]
    otopic=ctx["otopic"]
    for entry in states:
        key, value, stability=entry
        if use_kv:
            m={"topic":otopic, ctx["key"]: key, ctx["value"]: value, "stable": stability}
        else:
            m={"topic":otopic, "key": key, "value": value, "stable": stability}
        stdout(m)


#######################################################################################
#
#
#




"""
    Created on 2012-01-19
    @author: jldupont
"""
import logging, yaml

from jlddk.tools_sys import prepare_mod, raise_if_not_ok
from jlddk.tools_os import resolve_path, file_contents

from agents.exc import *

def run( path_config=None
         #,args=[]
        ,**_kw
        ):
    
    code, maybe_cpath=resolve_path(path_config)
    raise_if_not_ok(code, "Can't resolve configuration file path")
    
    code, maybe_contents=file_contents(maybe_cpath)
    raise_if_not_ok(code, "Can't resolve load configuration file")
    
    logging.info("Loading configuration file...")
    try:
        cdata=yaml.load(maybe_contents)
    except Exception,e:
        raise Exception("Problem with YAML config file: %s" % e)
    
    logging.info("Validating configuration file...")
    agents=cdata.get("agents", None)
    if agents is None:
        raise Exception("No 'agents' list found...")
    
    entries=[]
    for agent_name in agents:
        logging.info("Preparing agent '%s'" % agent_name)
        module_name=agents[agent_name]
        mod=prepare_mod(module_name)
        try:
            vtable=getattr(mod, "__vtable__")
        except:
            raise Exception("Can't find '__vtable__' in agent's module")
        entries.append((agent_name, module_name, vtable, mod))
        
            
        
    logging.info("Loaded %s agents" % len(agents))
    
    mqueue=[{"topic": "init", "ctx": cdata},]
    while True:
        try:     
            msg=mqueue.pop(0)
        except:
            raise Exception("No more messages in queue")
        
        try:    topic=msg["topic"]
        except: 
            raise Exception("Message is missing 'topic' field...")
        
        ### broadcast the msg to all agents
        for agent_name, module_name, vtable, _module in entries:
            fnc=vtable.get(topic, None)
            if fnc is None:
                fnc=vtable.get("*", None)
                if fnc is None:
                    continue
            try:
                fnc(mqueue, msg)
            except ExcInfo, e:
                logging.info("(%s:%s) %s" % (agent_name, topic, e))
            except ExcWarn, e:
                logging.warning("(%s:%s) %s" % (agent_name, topic, e))
            except ExcErr,e:
                logging.error("(%s:%s) %s" % (agent_name, topic, e))
            except ExcCrit,e:
                raise               
            except ExcQuit:
                raise KeyboardInterrupt()
            except Exception,e:
                logging.error("(%s:%s) %s" % (agent_name, topic, e))
            
        
        
    


"""
    Created on 2012-01-27
    @author: jldupont
"""
import sys, os, logging
from jlddk.tools_os import atomic_write

def maybe_log(verbose, msg):
    if verbose:
        logging.info(msg)
    

def run(path_dest=None,
        ignore_fault=False,
        verbose=False,
        ext=None
        ,keep_key=None
        ,**_
        ):
    
    maybe_log(verbose, "Resolving destination path: %s" % path_dest)
    try:
        apath=os.path.abspath(path_dest)
        path=apath.strip("\"'")
        path=os.path.expanduser(os.path.expandvars(path))
    except:
        raise Exception("Can't resolve path")

    ppid=os.getppid()        
    logging.info("Process pid: %s" % os.getpid())
    logging.info("Parent pid: %s" % ppid)
    logging.info("Starting loop...")
    while True:
        if os.getppid()!=ppid:
            logging.warning("Parent terminated... exiting")
            break
            
        try:
            iline=sys.stdin.readline().strip(" \n")
            if len(iline)==0:
                continue
        except:
            raise Exception("Exiting... probably broken pipe")
        
        try:
            key, contents=iline.split("\t", 1)
        except Exception, e:
            if not ignore_fault:
                raise Exception("Invalid input format: %s" % e)
            continue
        
        maybe_log(verbose, "Received input with key '%s'" % key)
        
        if ext is not None:
            dpath=os.path.join(path, key, ext)
        else:
            dpath=os.path.join(path, key)
            
        if keep_key:
            contents="%s\t%s" % (key, contents)
            
        maybe_log(verbose, "Writing to: %s" % dpath)
        result, _=atomic_write(dpath, contents)
        if not result.startswith("ok"):
            if not ignore_fault:
                raise Exception("Error writing output file '%s':%s" % (dpath, _))
        
        


"""
    Created on 2012-01-19
    @author: jldupont
"""
import logging, os, time

from jlddk.tools_sys import prepare_callable, stdout, stdout_flush

def run( module_name=None
        ,function_name=None
        ,fargs=None
        ,polling_interval=None
        ,**_kw
        ):
    
    logging.debug("Preparing callable...")
    
    try:
        _mod, fn=prepare_callable(module_name, function_name)
    except:
        raise Exception("Can't prepare callable: %s.%s" % (module_name, function_name))
    
    ppid=os.getppid()
    logging.info("Process pid: %s" % os.getpid())
    logging.info("Parent pid : %s" % ppid)
    logging.info("Starting loop...")
    while True:
        if os.getppid()!=ppid:
            logging.warning("Parent terminated... exiting")
            break

        try:
            result=fn(*fargs)
        except Exception, e:
            logging.warning("Problem with callable: %s" % str(e))
        
        if result is not None:
            stdout(result)
            stdout_flush()
            
        logging.debug("...sleeping for %s seconds" % polling_interval)
        time.sleep(polling_interval)
        



"""
    Created on 2012-01-27
    @author: jldupont
"""
import os, glob, logging
from jlddk.tools_os import file_contents, touch

def maybe_log(verbose, msg):
    if verbose:
        logging.info(msg)
    

def run(path_source=None,
        path_done=None,
        check_done=None,
        gen_done=None,  
        ignore_fault=False,
        verbose=False,
        rm_ext=False
        ,**_
        ):
    
    if check_done and path_done is None:
        raise Exception("Must specify 'done path'")
    
    maybe_log(verbose, "Resolving path for: %s" % path_source)
    try:
        apath=os.path.abspath(path_source)
        spath=apath.strip("\"'")
        spath=os.path.expanduser(os.path.expandvars(spath))
    except:
        raise Exception("Can't resolve path")

    dpath=None
    if check_done:
        maybe_log(verbose, "Resolving path for done files: %s" % path_done)
        try:
            dpath=os.path.abspath(path_done)
            dpath=dpath.strip("\"'")
            dpath=os.path.expanduser(os.path.expandvars(dpath))
        except:
            raise Exception("Can't resolve path")
        

    try:    
        files=glob.glob(os.path.join(spath, "*"))
    except:
        raise Exception("Can't fetch files")
    
    ## strip directories
    files=filter(os.path.isfile, files)
    
    for _file in files:
        
        if _file.endswith(".done"):
            continue
        
        bn=os.path.basename(_file)
        
        if check_done:
            pbn=os.path.join(dpath, bn)
            exists=os.path.exists(pbn+".done")
            if exists:
                maybe_log(verbose, "File '%s' already processed... skipping" % bn)
                continue
            
        try:
            maybe_log(verbose, "Processing '%s'" % bn)
            process_file(verbose, ignore_fault, rm_ext, _file)
        except Exception,e:
            if not ignore_fault:
                raise Exception("Can't process '%s': %s" % (bn, e))
            
        if gen_done:
            done_file=os.path.join(dpath, bn)
            result, _=touch(done_file+".done")
            if not result.startswith("ok"):
                if not ignore_fault:
                    raise Exception("Can't write 'done' file in 'done' path")
            else:
                maybe_log(verbose, "Wrote 'done' file for: %s" % bn)
    
    
def process_file(verbose, ignore_fault, rm_ext, _file):
    
    result, contents=file_contents(_file)
    if not result.startswith("ok"):
        if not ignore_fault:
            raise Exception("Can't read file: %s" % _file)
    
    try:    
        maybe_log(verbose, "Formatting file contents")
        contents=contents.replace('\t', ' ')
        lines=contents.split("\n")
        big_line='\t'.join(lines)
        
        name=os.path.basename(_file)
        if rm_ext:
            name, _=os.path.splitext(name)
        whole_line="%s\t%s" % (name, big_line)
    except:
        raise Exception("Can't format text file")
    
    maybe_log(verbose, "Outputting file")
    print whole_line
    
        
        
    
    


"""
    Created on 2012-01-27
    @author: jldupont
"""
import logging, os, signal
from time import sleep
import psutil


def run(prefix=None, polling_interval=None
        ,force_kill=None
        ,username=None
        ,ppid=None
        ,**_
        ):

    def prefix_match(entry):
        if prefix is None:
            return True
        
        bns=map(os.path.basename, entry.cmdline)
        flist=filter(lambda x:x.startswith(prefix), bns)
        return len(flist)>0
    

    def filtre(pentry):
        user_match=(username is None) or (username==pentry.username)
        name_match=(prefix is None)   or (pentry.name.startswith(prefix))
        ppid_match=(ppid is None)     or (pentry.ppid==ppid)
        pmatch=prefix_match(pentry)
        
        return ppid_match and (name_match or pmatch) and user_match


    ppid=os.getppid()        
    logging.info("Process pid: %s" % os.getpid())
    logging.info("Parent pid:  %s" % ppid)
    logging.info("Starting loop...")
    while True:
        if os.getppid()!=ppid:
            logging.warning("Parent terminated... exiting")
            break
        
        #deprecated in psutil
        #plist=psutil.get_process_list()
        #flist=filter(filtre, plist)
        
        for p in psutil.process_iter():
            
            if filtre(p):                
                cmdline=p.cmdline
                pid=p.pid
                user=p.username
                
                details="pid '%s' '%s' : %s" % (pid, user, cmdline)
                
                if not force_kill:
                    logging.info("Would kill %s" % details)
                else:
                    try:
                        os.kill(p.pid, signal.SIGTERM)
                        logging.info("Killed %s" % details)
                    except:
                        try:
                            os.kill(p.pid, signal.SIGKILL)
                            logging.info("Killed %s" % details)
                        except:
                            logging.warning("Couldn't kill %s" % details)
        
        logging.debug("Sleeping...")
        sleep(polling_interval)





"""
    Created on 2012-01-27
    @author: jldupont
"""
import logging, sys, os, json
from tools_os import atomic_write, can_write
from tools_web import fetch


def run(dest_path=None 
        ,**_):
    
    if dest_path is not None:
        if not os.path.isdir(dest_path):
            raise Exception("Expecting a valid destination path '%s'" % dest_path)

    ppid=os.getppid()            
    logging.info("Process pid: %s" % os.getpid())
    logging.info("Parent pid : %s" % ppid)
    logging.info("Starting loop...")
    while True:
        if os.getppid()!=ppid:
            logging.warning("Parent terminated... exiting")
            break
        
        iline=sys.stdin.readline().strip()
        
        #################### VALIDATE
        ## if we received two strings on the same line:  url  dst_path
        bits=iline.split(" ")
        l=len(bits)
        if l > 2 or l==0:
            logging.error("Invalid input line: %s" % iline)
            continue
        
        url=bits[0]
        bn=os.path.basename(url)
        
        if len(bits)==2:
            path=bits[1]
        else:
            if dest_path is not None:
                path=os.path.join(dest_path, bn)
            else:
                logging.warning("Didn't receive 'dest_path' from stdin and none specified on command line...")
                continue
            
        ####### WRITE CAPABILITY VERIFICATION
        code, result=can_write(path)
        if not code.startswith("ok") or not result:
            logging.warning("Won't be able to write to path '%s'... skipping download" % path)
            continue
            
        ####### DOWNLOAD
            
        code, (http_code, headers, data)=fetch(url)
        if not code.startswith("ok"):
            logging.warning("Error attempting to download: %s" % url)
            continue
        
        try:     http_code=int(http_code)
        except:  pass
        
        if http_code!=200:
            logging.warning("Can't fetch url '%s', http response code: %s" % (url, http_code))
            continue
                        
        code, msg=atomic_write(path, data)
        if not code.startswith("ok"):
            raise Exception("Can't write to file '%s': %s" % (path, msg))
        
        ctx={
             "dest_filename": path
             ,"url": url
             ,"http_code": http_code
             ,"headers": headers
             }
        
        try:    sys.stdout.write(json.dumps(ctx)+"\n")
        except: 
            raise Exception("Exiting... probably broken pipe")
            
        

            


"""
    Created on 2012-01-28
    @author: jldupont
"""
import os,sys,logging, hashlib, random
from logging.handlers import SysLogHandler

def setloglevel(level_name):
    """
    >>> import logging
    >>> setloglevel("info")
    >>> logging.debug("test")
    
    """
    try:
        ll=getattr(logging, level_name.upper())
        logger=logging.getLogger()
        logger.setLevel(ll)
    except:
        raise Exception("Invalid log level name: %s" % level_name)

class FilterDuplicates(logging.Filter):
    """
    Everything before the ':' marker is considered to be the "signature" for a log event
    
    - All DEBUG level log go through
    - All "progress" report go through
    - All messages with a ":" separator (for contextual info) are passed
    - Other messages are filtered for duplicates 
    """
    occured=[]
    
    def filter(self, record):
        if record.levelname=="DEBUG":
            return 1
        
        msg=record.getMessage()
        
        if msg.startswith("progress") or msg.startswith("Progress") or msg.startswith("PROGRESS"):
            return 1
        
        try:
            bits=msg.split(":")
            if len(bits)>1:
                return 1
            
            signature_hash=hashlib.md5(msg).hexdigest()
            if signature_hash in self.occured:
                return 0
            
            self.occured.append(signature_hash)
            record.msg="*"+msg
        except Exception,e:
            print e
            ### let it pass...
            pass
        
        return 1
        


def enable_duplicates_filter():
    logger=logging.getLogger()
    logger.addFilter(FilterDuplicates())

########################################################
# Syslog support
#
def _get_fname():
    name=os.path.basename(sys.argv[0])
    cmdname=os.environ.get("CMDNAME", name)
    return "%-12s" % cmdname


FORMAT='%(asctime)s - '+_get_fname()+' - %(levelname)s - %(message)s'
     

def setup_basic_logging():    
    logging.basicConfig(level=logging.INFO, format=FORMAT)
    

def setup_syslog():

    slogger=SysLogHandler()
    
    slogger.setLevel(logging.INFO)
    formatter = logging.Formatter(FORMAT)
    slogger.setFormatter(formatter)
    
    logger=logging.getLogger()
    logger.addHandler(slogger)


if __name__=="__main__":
    import doctest
    doctest.testmod()
    


"""
    Created on 2012-01-27
    @author: jldupont
"""
import logging, sys, json, os, select, time
from tools_os import resolve_path, getsubdirs
from tools_func import simple_transition_manager, transition_manager 

from pyfnc import partial

def stdout(jo):
    try:    
        sys.stdout.write(json.dumps(jo)+"\n")
        sys.stdout.flush()
    except:
        raise Exception("Exiting... probably broken pipe")

def run( path=None
        ,polling_interval=None
        ,topic=None
        ,always=None
        ,**_
        ):

    code, path=resolve_path(path)
    if not code.startswith("ok"):
        raise Exception("Can't resolve path...: %s" % path)

    def loginfo(path, state, *_):
        logging.info("Path state '%s': %s" % (path, state))

    ctx={"topic": topic, "path": path, "always":always}
    ctx["_path"]={
                  "previous": "ok"
                  ,"ch": partial(loginfo, path)
                  }
    tm=transition_manager(ctx)
    

    ppid=os.getppid()
    logging.info("Process pid: %s" % os.getpid())
    logging.info("Parent pid : %s" % ppid)
    logging.info("Starting loop...")    
    while True:
        if os.getppid()!=ppid:
            logging.warning("Parent terminated... exiting")
            break

        code, maybe_subdirs=getsubdirs(path)
        tm.send(("_path", code))
        if code.startswith("ok"):
            process(ctx, maybe_subdirs)
        
        start_time=time.time()
        while True:
            ir, _w, _e=select.select([sys.stdin], [], [], polling_interval)
            if len(ir):
                iline=sys.stdin.readline()
                sys.stdout.write(iline)
                
            elapsed_time = time.time() - start_time
            if elapsed_time > polling_interval:
                break


def getmtime(path):
    return (path, os.path.getmtime(path))

def getchanges(ctx, entry):
    path, mtime=entry
    result=simple_transition_manager(ctx, path, mtime)
    return (path, mtime, result)

def process(ctx, subdirs):
    topic=ctx["topic"]
    path=ctx["path"]
    always=ctx["always"]
    
    minfo=map(getmtime, subdirs)
    
    entries=map(partial(getchanges, ctx), minfo)
    
    for entry in entries:
        path, mtime, result=entry
        maybe_tr, _=result
        if maybe_tr=="tr" or always:
            output(topic, path, mtime)


def output(topic, path, mtime):
    
    d={
       "topic": topic
       ,"path": path
       ,"mtime": mtime
       }
    stdout(d)

#
#


"""
    Created on 2012-01-27
    @author: jldupont
"""
import logging, sys, json, os
from time import sleep

ONE_SECOND=1

def stdout(jo):
    sys.stdout.write(json.dumps(jo)+"\n")
        

def run(topic_name=None
        ,separate_msg_marker=False
        ,suppress_second_marker=False
        ,**_
        ):

    ### elapsed
    d={"topic": topic_name
       ,"sec": 0
       ,"min": 0
       ,"hour": 0
       ,"day": 0
       ,"min_marker": False
       ,"hour_marker": False
       ,"day_marker": False
    }

    ppid=os.getppid()        
    logging.info("Process pid: %s" % os.getpid())
    logging.info("Parent pid: %s" % ppid)
    logging.info("Starting loop...")
    
    while True:
        if os.getppid()!=ppid:
            logging.warning("Parent terminated... exiting")
            break

        d["min_marker"]=False
        d["hour_marker"]=False
        d["day_marker"]=False
        
        sec=d["sec"]+1
        d["min_marker"]=(sec==60)
        d["sec"]=sec % 60
        
        if d["min_marker"]:
            _min=d["min"]+1
            d["hour_marker"]=(_min==60)
            d["min"]=_min % 60
            
            if d["hour_marker"]:            
                hour=d["hour"]+1
                d["day_marker"]=(hour==24)
                d["hour"]=hour % 24
                
                if d["day_marker"]:                    
                    d["day"]=d["day"]+1
        
        #logging.debug(" sec(%s) min_marker(%s) hour_marker(%s) day_marker(%s)" % (sec, min_marker, hour_marker, day_marker))
        if suppress_second_marker:
            if not d["min_marker"] and not d["hour_marker"] and not d["day_marker"]:
                sleep(ONE_SECOND)
                continue
               
        try:
            stdout(d)
            
            if separate_msg_marker:
                if d["min_marker"]:
                    stdout({"topic":"min_marker"})
                if d["hour_marker"]:
                    stdout({"topic":"hour_marker"})
                if d["day_marker"]:
                    stdout({"topic":"day_marker"})
        except:
            raise Exception("Exiting... probably broken pipe")
        
        sleep(ONE_SECOND)





"""
    @author: Jean-Lou Dupont
"""

import sys,os
from tools_logging import setup_basic_logging


## force stdout flushing
sys.stdout = os.fdopen(sys.stdout.fileno(), 'w', 0)
sys.stderr = os.fdopen(sys.stderr.fileno(), 'w', 0)

setup_basic_logging()



"""
    Created on 2012-01-27
    @author: jldupont
"""
import sys, os, logging
from jlddk.tools_os import atomic_write, mkdir_p, touch

def run(
        path_dest=None
        ,append_ext=None
        ,disable_pass_through=False
        ,**_
        ):
    
    if append_ext is not None:
        append_ext=append_ext.strip(".'\"")
        logging.info("Using '%s' as extension" % append_ext)
    
    logging.info("Resolving destination path: %s" % path_dest)
    try:
        apath=os.path.abspath(path_dest)
        dpath=apath.strip("\"'")
        dpath=os.path.expanduser(os.path.expandvars(dpath))
    except:
        raise Exception("Can't destination resolve path")
    
    logging.info("Resolved destination path: %s" % dpath)

    logging.info("Creating destination, if required...")
    mkdir_p(dpath)

    ppid=os.getppid()        
    logging.info("Process pid: %s" % os.getpid())
    logging.info("Parent pid:  %s" % ppid)
    logging.info("Starting loop...")
    while True:
        if os.getppid()!=ppid:
            logging.warning("Parent terminated... exiting")
            break
            
        try:
            iline=sys.stdin.readline().strip(" \n")
            if len(iline)==0:
                continue
        except:
            raise Exception("Exiting... probably broken pipe")
        
        if not disable_pass_through:
            sys.stdout.write(iline+"\n")
            
        if append_ext is not None:
            filename=iline+".%s" % append_ext
        else:
            filename=iline
            
        path=os.path.join(dpath, filename)
        
        code, _=touch(path)
        if not code.startswith("ok"):
            logging.error("Can't touch '%s'" % path)

            
        


"""

    Ready --> Pending
    Ready --> Error
    Ready --> Stopped
    
    Pending --> Ready
    Pending --> Error
    Pending --> Stopped
    
    Error   --> Ready
    Error   --> Stopped

    Stopeed --> 

    Q:What if a task takes too much time to complete?
                    doesn't send a "done" or "error" message?
                    
    A: don't have any control over here.

    Input Messages:
    
        x_done
        x_error
        x_worker
        
    Output Messages:
    
        x_timeout
        ?x_task


    Created on 2012-01-27
    @author: jldupont
"""
import logging, sys, json, os
import types

from tools_sys import BrokenPipe
from tools_func import transition_manager

from pyfnc import pattern, patterned, dic

MTYPES=["task", "done", "error", "worker", "none"]

def run(args
        ,**_ 
        ):
    
    def dp(msg):
        def _(state):
            logging.debug(msg % state)
        return _
        
    
    tmctx={ 
            "worker": { "ch": dp("Worker is: %s"), "up": dp("Worker is: %s") }
           #,"state":  { "up": dp, "down": dp }
           }
    
    ctx=dict(args)
    ctx["_tm"]=transition_manager(tmctx)
         
    ppid=os.getppid()    
    logging.info("Process pid: %s" % os.getpid())
    logging.info("Parent pid : %s" % ppid)
    logging.info("Starting loop...")
    while True:
        if os.getppid()!=ppid:
            logging.warning("Parent terminated... exiting")
            break

        try:
            iline=sys.stdin.readline().strip(" \n")
        except KeyboardInterrupt:
            raise
        except:
            raise BrokenPipe("Broken Pipe...")

        #logging.debug("Received: %s" % iline)

        if iline=="":
            continue

        try:
            jso=json.loads(iline)
        except:
            try:    logging.debug("Can't JSON decode: %s" % iline)
            except: pass
            continue            
        
        #logging.debug("Received: %s" % jso)

        try:
            topic=jso["topic"]
        except:
            logging.warning("Can't find 'topic' member in json object: %s" % jso)
            continue
        
        code, msg=process(ctx, topic, jso)
        hcode(code, msg)

############################################################################################

@pattern("ok", any)
def hcode_ok(_, msg):
    pass

@pattern("debug", any)
def hcode_debug(_, msg):
    logging.debug(msg)

@pattern(any, any)
def hcode_any(code, msg):
    raise Exception("Unsupported code: %s" % code)

@patterned
def hcode(code, msg): pass


############################################################################################

def process(ctx, topic, jso):
    if topic==ctx["clock_topic"]:
        return hclock(ctx, jso)

    try:    
        task_type, mtype=topic.split("_")
        question=task_type.startswith("?")
        task_type=task_type.lstrip("?")
    except: 
        return ("debug", "invalid topic name: %s" % topic)
    
    logging.debug("Process: %s %s %s" % (question, task_type, mtype))
    
    task_types=ctx["task_types"]
    if task_type not in task_types:
        return ("debug", "nothing to do: %s (%s)" % (topic, task_type))
    
    if mtype not in MTYPES:
        return ("debug", "mtype not interesting: %s" % mtype)
    
    return handle(ctx, question, task_type, mtype)

###########################################################################################    

def hclock(ctx, _jso):
    """
    Have some timeouts expired?
    
    _$ttype: { "timeout": $timeout_left_in_seconds }
    """
    #ctx["_clock"]=jso
    tm=ctx["_tm"]
    
    timeout_state_report=ctx["timeout_state_report"]
    
    ### go through all 'ttype' and decrease their timeout
    ttypes=ctx["task_types"]
    for ttype in ttypes:
        
        ### state timeout
        timeout=tget(ctx, ttype, "timeout", 1)
        timeout=(timeout-1) if timeout>0 else 0
        tset(ctx, ttype, "timeout", timeout)
        
        ### state reporting
        ###
        timeout_report=tget(ctx, ttype, "report_state", timeout_state_report)
        
        if timeout_state_report!=0:
            if timeout_report==0:
                current=tget(ctx, ttype, "state", "ready")
                send_msg(False, ttype, "state", {"state": current})
        
        timeout_report=(timeout_report-1) if timeout_report>0 else timeout_state_report
        tset(ctx, ttype, "report_state", timeout_report)
        
        
        ### worker keep-alive timeout
        max_timeout_worker=ctx["max_timeout_worker"]
        wtimeout=tget(ctx, ttype, "timeout_worker", max_timeout_worker)
        wtimeout=(wtimeout-1) if wtimeout>0 else 0
        tset(ctx, ttype, "timeout_worker", wtimeout)
        
        if wtimeout==0:
            tm.send(("worker", "stopped"))
            tset(ctx, ttype, "state", "stopped")

    ### now, compute the "next" state
    for ttype in ttypes:
        
        current=tget(ctx, ttype, "state", "ready")
        new=compute(ctx, current, ttype)
        tset(ctx, ttype, "state", new)
        
        #tm.send(("state", new))
    
    return ("ok", None)


###########################################################################################
###
###  STATE-MACHINE
###

@pattern(dict, "ready", str)
def compute_1(ctx, _, ttype):
    """
    Timeout expired? we are ready to ask for a new task to be performed
    """
    timeout=tget(ctx, ttype, "timeout", 0)
    if timeout==0:
        
        ###  ?x_task
        ###
        send_msg(True, ttype, "task", {})
        max_pending=ctx["max_pending"]
        tset(ctx, ttype, "timeout", max_pending)
        return "pending"
    
    return "ready"


@pattern(dict, "error", str)
def compute_2(ctx, _, ttype):
    """
    timeout expired?  we can get out of "error" state
    """
    timeout=tget(ctx, ttype, "timeout", 0)
    if timeout==0:
        return "ready"
    return "error"

@pattern(dict, "pending", str)
def compute_3(ctx, _, ttype):
    """
    timeout expired? need to advise that there is potentially a problem with the worker...
    """
    timeout=tget(ctx, ttype, "timeout", 0)
    if timeout==0:
        
        ###  x_timeout
        ###
        send_msg(False, ttype, "timeout", {})
        return "ready"
    
    return "pending"


@pattern(dict, "stopped", str)
def compute_4(ctx, _, ttype):
    """
    worker now available? exit "stopped" state
    """
    timeout_worker=tget(ctx, ttype, "timeout_worker", 0)
    if timeout_worker > 0:
        tset(ctx, ttype, "timeout", 0)
        return "ready"
    
    return "stopped"
    

@pattern(any, any, any)
def compute_any(ctx, current, ttype):
    raise Exception("Unknown state: %s" % current)


@patterned
def compute(ctx, current, ttype): pass


###########################################################################################
###
### SUPPORT FUNCTIONS

def tgetall(ctx, ttype, default={}):
    return ctx.get("_"+ttype, default)

def tget(ctx, ttype, key, default=None):
    _ttype=ctx.get("_"+ttype, {})
    return _ttype.get(key, default)

def tset(ctx, ttype, key, value):
    _ttype=ctx.get("_"+ttype, {})
    _ttype[key]=value
    ctx["_"+ttype]=_ttype


def send_msg(question, ttype, mtype, msg_dic):
    try:
        q="?" if question else ""
        topic=q+ttype+"_"+mtype
        
        m=dic({"topic": topic}).update(msg_dic)
        
        sys.stdout.write(json.dumps(m)+"\n")
    except:
        pass


###########################################################################################
###
### MESSAGE HANDLING
###
@pattern(dict, False, types.UnicodeType, "none")
def handle_none(ctx, _question, ttype, _):
    """
    No tasks are available...
    No need to be sure a worker is available... 
        saves from implementing a loop for sending "dl_worker" for nothing...
        
    Pretend worker is OK - we don't really care at this point.
    """
    max_timeout_worker=ctx["max_timeout_worker"]
    tset(ctx, ttype, "timeout_worker", max_timeout_worker)
    
    tset(ctx, ttype, "state",   "ready")
    return ("ok", None)


@pattern(dict, False, types.UnicodeType, "worker")
def handle_worker(ctx, _question, ttype, _):
    """
    Keep alive message related to a specific task-type
    
    Reset timeout associated with task-type
    When this counter reaches 0 ===>  worker is considered off-line
    """
    max_timeout_worker=ctx["max_timeout_worker"]
    tset(ctx, ttype, "timeout_worker", max_timeout_worker)
    logging.debug("Worker keep-alive: %s" % ttype)
    
    tm=ctx["_tm"]
    tm.send(("worker", "started"))
    return ("ok", None)
    

@pattern(dict, False, types.UnicodeType, "error")
def handle_error(ctx, _question, ttype, _):
    """
    Specific "task type" in error ==> handle timeout
    
    Double the current timeout (min of 1 second) until max_timeout_error is reached
    """
    max_timeout=ctx["max_timeout_error"]
    timeout=tget(ctx, ttype, "timeout", 1)
    
    timeout=min(2, timeout)
    timeout=min(timeout*2, max_timeout)
    
    tset(ctx, ttype, "timeout", timeout)
    tset(ctx, ttype, "state",   "error")
    
    return ("ok", None)

@pattern(dict, False, types.UnicodeType, "task")
def handle_task(ctx, _question, ttype, _):
    """
    Cool... there is a dispatcher but we don't know if the task is going to get done...
    """
    return ("ok", None)

@pattern(dict, False, types.UnicodeType, "done")
def handle_done(ctx, _question, ttype, _):
    """
    If we receive this sort of message, 
    it probably means there is no more error related to this task-type
    
    Clear outstanding task
    """
    tset(ctx, ttype, "state", "ready")
    
    ### need to wait a bit between tasks...
    wait=ctx["wait"]
    tset(ctx, ttype, "timeout", wait)
    return ("ok", None)

@pattern(any, any, any, any)
def handle_any(_ctx, question, ttype, mtype):
    """ probably an error to end up here """
    q="?" if question else ""
    logging.debug(type(_ctx))
    logging.debug(type(question))
    logging.debug(type(ttype))
    logging.debug(type(mtype))
    return ("debug", "nothing to do with: %s%s_%s" % (q, ttype, mtype))


@patterned
def handle(ctx, question, task_type, mtype): pass


########################################################################3




#!/usr/bin/env python
"""
    Jean-Lou Dupont's general web related scripts/robots
    
    Created on 2012-01-19
    @author: jldupont
"""
__author__  ="Jean-Lou Dupont"
__version__ ="0.7.14"


from distutils.core import setup
from setuptools import find_packages

DESC="""
Overview
--------

This package contains a collection of 'robots'.
Each 'robot' is implemented with logging on stderr and processing result is output on stdout as either string or JSON.

Robots
------

* jldstrpkg: stdin --> formatter --> stdout
* jldwebscraper : extract anchor links from a web page
* jldfilter : filter stdin through a Python module, output on stdout
* jldcomp : compare 2 filesystem paths for difference in files
* jldfcomp : compare globs of files
* jldcat: 'cat' files to stdout, 1 per line with format 'filename\tcontents\n'
* jlduncat: Receive from stdin "key\tcontents\n" and write to key[.ext]
* jldjsoncat : 'cat' files to stdout by encapsulating them in JSON objects
* jldwebdl : gated web page download, source links contained in file system path
* jldinotify : path change notification over JSON/stdout
* jldfetcher : web page fetcher which takes instructions from stdin
* jldfilelist : list path files using include/exclude filter, JSON/stdout output
* jldclock : 1 second interval clock with min,hour,day markers
* jldpclean : kills processes left to pid=1
* jldtaskctl : task controller
* jldostr : outputs a string on stdout @ regular interval, pass-through for stdin-->stdout
* jldstatsubdirs: periodic status information on sub-dirs of path (JSON output & pass-through stdin-->stdout)
* jlddebouncer: debounces {key:value} pairs (JSON output & pass-through stdin-->stdout)
* jldarun : runs "agents"
* jldrun : periodically run a python callable
* jldtouch: receive filename from stdin, 'touch' filename in destination path
* jldsplitter: a semi customizable file splitter

Configuration
-------------

Can be performed through options on the command line or using a file (use a leading `@`).
"""

setup(name=         'jlddk',
      version=      __version__,
      description=  'Collection of robots',
      author=       __author__,
      author_email= 'jl@jldupont.com',
      url=          'http://www.systemical.com/doc/opensource/jlddk',
      package_dir=  {'': "src",},
      packages=     find_packages("src"),
      scripts=      ['src/scripts/jldwebscraper',
                     'src/scripts/jldfilter',
                     'src/scripts/jldwebdl',
                     'src/scripts/jldcat',
                     'src/scripts/jlduncat',
                     'src/scripts/jldjsoncat',
                     'src/scripts/jldcomp',
                     'src/scripts/jldfcomp',
                     'src/scripts/jldinotify',
                     'src/scripts/jldfetcher',
                     'src/scripts/jldfilelist',
                     'src/scripts/jldclock',
                     'src/scripts/jldpclean',
                     'src/scripts/jldtaskctl',
                     'src/scripts/jldostr',
                     'src/scripts/jldstatsubdirs',
                     'src/scripts/jlddebouncer',
                     'src/scripts/jldarun',
                     'src/scripts/jldrun',
                     'src/scripts/jldtouch',
                     'src/scripts/jldsplitter',
                     ],                     
      zip_safe=False
      ,install_requires=["pyfnc >= 0.1.0"
                         ,"psutil"
                         ,"pyinotify"
                         ,"pyyaml >= 3.10"
                         ]
      ,long_description=DESC
      )

#############################################

f=open("latest", "w")
f.write(str(__version__)+"\n")
f.close()



"""
    Created on 2012-02-17
    @author: jldupont
    
"""
import logging, sys
import time, select

from tools_sys import BrokenPipe

def run(ostr=None,
        polling_interval=None
        ,**_
        ):
        
    logging.info("Starting... ")
    while True:
        
        try:
            sys.stdout.write(ostr+"\n")
        except:
            raise BrokenPipe("broken pipe...")

        logging.debug("...waiting for %s seconds (max)" % polling_interval)
        
        ### Implement a "pass-through" for stdin --> stdout
        ###  whilst also handling a maximum time-out
        start_time=time.time()
        while True:
            ir, _w, _e=select.select([sys.stdin], [], [], polling_interval)
            if len(ir):
                iline=sys.stdin.readline()
                sys.stdout.write(iline)
                
            elapsed_time = time.time() - start_time
            if elapsed_time > polling_interval:
                break



"""
    Created on 2012-01-27
    @author: jldupont
"""
import os
import glob


def run(list_matches=None,
        list_diffs=None,
        list_sets=None,
        just_basename=None,
        just_filename=None,
        path_a=None,
        path_b=None,
        verbose=None
        ,**_
        ):

    try:
        ga=glob.glob(path_a)
    except:
        raise Exception("Pattern for path 'a' is invalid")

    try:
        gb=glob.glob(path_b)
    except:
        raise Exception("Pattern for path 'b' is invalid")
    

    if just_basename:
        ga=map(os.path.basename, ga)
        gb=map(os.path.basename, gb)
        
    if just_filename:
        ga=map(os.path.splitext, ga)
        ga=map(lambda t:t[0], ga)
        gb=map(os.path.splitext, gb)
        gb=map(lambda t:t[0], gb)
        
            
    sa=set(ga)
    sb=set(gb)
    
    if list_matches:
        out(sa.intersection(sb), verbose, '== ')
        
    if list_diffs:
        out(sa.symmetric_difference(sb), verbose, '!= ')
        
    if list_sets:
        out(sa, verbose, "pa: ")
        out(sb, verbose, "pb: ")

def out(liste, verbose, prefix):
    if not verbose:
        prefix=''
    for item in liste:
        print "%s%s" % (prefix, item)
    


"""
    Created on 2012-07-21
    @author: jldupont
"""
import logging


def init(_q, *p):
    print "test2.init", p

def tick(_q, _):
    logging.info("test2: tick!")



__vtable__={"init": init, "tick": tick}    